//
// Buffer.cpp
//  Buffer managment for HAIL
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <math.h>

#define for if(0);else for

#include "Buffer.h"
#include "ATIExtensions.h"
#include "PBuffer.h"

#define MAX_ATTRIBS 32
#define MAX_FORMATS 256

#define WIDTH 1024
#define HEIGHT 1024



PBuffer* floatBuf = NULL;
PBuffer* usBuf = NULL;

static int bufferWidth;
static int bufferHeight;

static unsigned int utilityTex[MAX_TEXTURE];



//
//
////////////////////////////////////////////////////////////////////////////////
bool CreateBuffers()
{
   

   floatBuf = new PBuffer( WIDTH, HEIGHT, 128, true, false, false, false, false, true);
   CreateUtilityTextures();
   //wglMakeCurrent( hdc, rc);

   bufferWidth = WIDTH;
   bufferHeight = HEIGHT;
   return true;
}

//
//
////////////////////////////////////////////////////////////////////////////////
void DestroyBuffers()
{
   
   delete floatBuf;
}

//
//
////////////////////////////////////////////////////////////////////////////////
int GetBufferWidth()
{
   return bufferWidth;
}

//
//
////////////////////////////////////////////////////////////////////////////////
int GetBufferHeight()
{
   return bufferHeight;
}

//
//
////////////////////////////////////////////////////////////////////////////////
void BindBuffer( BufferType type)
{
   

   switch (type)
   {
      case FLOAT_BUFFER:
         floatBuf->Bind();
         break;
   };
}

//
//
////////////////////////////////////////////////////////////////////////////////
void UnbindBuffer(BufferType type)
{
   switch (type)
   {
      case FLOAT_BUFFER:
         floatBuf->Unbind();
         break;
   };
   //wglMakeCurrent( prevDC, prevRC);
}

//
//
////////////////////////////////////////////////////////////////////////////////
bool IsBufferSupported( BufferType type)
{
   bool ret;

   switch (type)
   {
      case UBYTE_BUFFER:
         ret = false;
         break;
      case USHORT_BUFFER:
         ret = false;
         break;
      case FLOAT_BUFFER:
         ret = floatBuf->IsValid();
         break;
   };

   return ret;
}

//
//
////////////////////////////////////////////////////////////////////////////////
void CreateUtilityTextures()
{
   float *fImage;
   GLushort *sImage;
   float x,y;

   BindBuffer( FLOAT_BUFFER);

   glGenTextures( MAX_TEXTURE, utilityTex);

   //these must match the order of the enum declaration


   //linear
   glBindTexture( GL_TEXTURE_2D, utilityTex[0]);
   sImage = new GLushort[256*2];

   for (int ii=0; ii<256; ii++)
   {
      x = (ii<128) ? (float)ii/127.0f : 1.0f - (float)(ii-128)/127.0f;
      sImage[ii*2] = (GLushort)(x*((1<<16)-1));
      sImage[ii*2+1] = (GLushort)((1.0f-x)*((1<<16)-1));
   }
   
   glTexImage2D( GL_TEXTURE_2D, 0, GL_LUMINANCE16_ALPHA16, 256, 1, 0, GL_LUMINANCE_ALPHA, GL_UNSIGNED_SHORT, sImage);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   delete []sImage;

   //cubic1 - 
   glBindTexture( GL_TEXTURE_2D, utilityTex[1]);
   fImage = new GLfloat[2048*4];

   float a = -0.5f;

   //this may be off center
   for (int ii=0; ii<2048; ii++)
   {
      x = (float)ii/2047.0f;

      y = 2.0f - x;
      y = fabsf(y);
      fImage[ii*4+0] = a*y*y*y - 5.0f*a*y*y + 8.0f*a*y - 4.0f*a;
      y = 1.0f - x;
      y = fabsf(y);
      fImage[ii*4+1] = (a+2.0f)*y*y*y - (a+3.0f)*y*y +1.0f;
      y = fabsf(x);
      fImage[ii*4+2] = (a+2.0f)*y*y*y - (a+3.0f)*y*y +1.0f;
      y = 1.0f + x;
      y = fabsf(y);
      fImage[ii*4+3] = a*y*y*y - 5.0f*a*y*y + 8.0f*a*y - 4.0f*a;
   }

   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, 2048, 1, 0, GL_RGBA, GL_FLOAT, fImage);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   delete []fImage;


   //scramble texture
   glBindTexture( GL_TEXTURE_2D, utilityTex[2]);
   fImage = new float[2048];
   float temp;

   for (int ii=0; ii<12; ii++)
   {
      for (int gg=0; gg<(2048>>ii); gg++)
      {
         fImage[gg] = (float)(gg+0.5f)/(float)((2048>>ii));
      }
      //apply scramble
      int jj = 0;
      for (int gg=0; gg<(2048>>ii); gg++)
      {
         if ( gg > jj)
         {
            temp = fImage[jj];
            fImage[jj] = fImage[gg];
            fImage[gg] = temp;
         }
         int m = (2048>>ii)>>1;
         while ((jj>=m) & (m >= 2))
         {
            jj -= m;
            m = m >>1;
         }
         jj += m;
      }
      glTexImage2D( GL_TEXTURE_2D, ii, GL_INTENSITY16, 2048>>ii, 1, 0, GL_LUMINANCE, GL_FLOAT, fImage);
   }

   delete []fImage;

   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_NEAREST);



   int texNum = 3;


   //fourier textures 
   for (int kk=1024,pow2=10; kk>8; kk>>=1, pow2--)
   {
      glBindTexture( GL_TEXTURE_2D, utilityTex[texNum++]);
      fImage = new float[kk*16*4];

      //zero the texture
      memset( fImage, 0, kk*16*4*sizeof(float));

      {
         float angle;
         float wre, wim, wpre, wpim, tempre, tempim;
         int jj;
         int N, half_N;
         float wtemp;

         N = 1;
         for (int ii=0; ii<pow2; ii++) //pow2 iterations
         {
            half_N = N;
            N <<= 1;
            angle = -2.0f * 3.1415927f / (float) N; //*dir
            wtemp = sinf( 0.5f * angle);
            wpre = -2.0f * wtemp *wtemp;
            wpim = sinf( angle);
            wre = 1.0f;
            wim = 0.0f;
            for (int offset=0; offset<half_N; offset++)
            {
               for (int gg=offset; gg<kk; gg+=N)
               {
                  jj = gg + half_N;
                  //store the coords in r and g
                  fImage[(ii*kk+jj)*4 + 0] = -((jj+0.5f)/(float)kk);
                  fImage[(ii*kk+jj)*4 + 1] = (gg+0.5f)/(float)kk;
                  fImage[(ii*kk+gg)*4 + 0] = (jj+0.5f)/(float)kk;
                  fImage[(ii*kk+gg)*4 + 1] = (gg+0.5f)/(float)kk;

                  //store W in b and a
                  fImage[(ii*kk+jj)*4 + 2] = wre;
                  fImage[(ii*kk+jj)*4 + 3] = wim;
                  fImage[(ii*kk+gg)*4 + 2] = wre;
                  fImage[(ii*kk+gg)*4 + 3] = wim;
               }
               wtemp = wre;
               wre = wtemp*wpre - wim*wpim + wre;
               wim = wim*wpre + wtemp*wpim + wim;
            }
         }
      }
      

      glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, kk, 16, 0, GL_RGBA, GL_FLOAT, fImage);

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

      delete []fImage;


      //inverse fourier texture (512)
      glBindTexture( GL_TEXTURE_2D, utilityTex[texNum++]);
      fImage = new float[kk*16*4];

      //zero the texture
      memset( fImage, 0, kk*16*4*sizeof(float));

      {
         float angle;
         float wre, wim, wpre, wpim, tempre, tempim;
         int jj;
         int N, half_N;
         float wtemp;

         N = 1;
         for (int ii=0; ii<pow2; ii++) //log2(512) iterations
         {
            half_N = N;
            N <<= 1;
            angle = -2.0f * 3.1415927f / (float) -N;
            wtemp = sinf( 0.5f * angle);
            wpre = -2.0f * wtemp *wtemp;
            wpim = sinf( angle);
            wre = 1.0f;
            wim = 0.0f;
            for (int offset=0; offset<half_N; offset++)
            {
               for (int gg=offset; gg<kk; gg+=N)
               {
                  jj = gg + half_N;
                  //store the coords in r and g
                  fImage[(ii*kk+jj)*4 + 0] = -(jj+0.5f)/(float)kk;
                  fImage[(ii*kk+jj)*4 + 1] = (gg+0.5f)/(float)kk;
                  fImage[(ii*kk+gg)*4 + 0] = (jj+0.5f)/(float)kk;
                  fImage[(ii*kk+gg)*4 + 1] = (gg+0.5f)/(float)kk;

                  //store W in b and a
                  fImage[(ii*kk+jj)*4 + 2] = wre;
                  fImage[(ii*kk+jj)*4 + 3] = wim;
                  fImage[(ii*kk+gg)*4 + 2] = wre;
                  fImage[(ii*kk+gg)*4 + 3] = wim;
               }
               wtemp = wre;
               wre = wtemp*wpre - wim*wpim + wre;
               wim = wim*wpre + wtemp*wpim + wim;
            }
         }
      }

      glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA_FLOAT32_ATI, kk, 16, 0, GL_RGBA, GL_FLOAT, fImage);

      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);

      delete []fImage;
   }

   UnbindBuffer(FLOAT_BUFFER);
}

//
//
////////////////////////////////////////////////////////////////////////////////
void BindUtilityTexture( UtilityTexture tex)
{
   glBindTexture( GL_TEXTURE_2D, utilityTex[tex]);
}