//
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include "HAIL.h"
#include "HAIL_private.h"

#include "Buffer.h"
#include "ATIExtensions.h"


#define for if(0);else for

static bool HAIL_initialized = false;

//
//
//////////////////////////////////////////////////////////////////////
bool HAIL::Init()
{
   if (HAIL_initialized)
   {
      return true;
   }

   SetupARBExtensions();
   SetupWGLExtensions();
   SetupATIExtensions();
   if ( (!GL_ARB_multitexture_Flag) || (!WGL_ARB_pbuffer_Flag) || (!GL_ATI_texture_float_Flag) || (!GL_ARB_fragment_program))
   {
      return false;
   }

   if ( !CreateBuffers())
   {
      return false;
   }

   HAIL_initialized = true;
   return true;
}

//
//
//////////////////////////////////////////////////////////////////////
void HAIL::Shutdown()
{
   //warning, this could orphan some buffers
   DestroyBuffers();

   HAIL_initialized = false;
}

//
//
//////////////////////////////////////////////////////////////////////
void HAIL::CreateShaderPreamble( strstream &str, int numInterp, int numTemps)
{
   //header
   str << "!!ARBfp1.0\n\n";

   //interpolators
   str << "# Interpolator declaration\n";
   for (int ii=0; ii<numInterp; ii++)
   {
      str << "ATTRIB t" << ii << " = fragment.texcoord[" << ii <<"];\n";
   }
   str << "\n";

   //temporaries
   str << "# Temp declartions\n";
   for (int ii=0; ii<numTemps; ii++)
   {
      str << "TEMP r" << ii <<";\n";
   }
   str << "\n";

   //output
   str << "# Output declarations\n";
   str << "OUTPUT res = result.color;\n\n";
}