//
//
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <gl/glut.h>
#include "glATI.h"
#include <stdio.h>
#include <stdlib.h>

#include <math.h>

#include "HAIL.h"
#include "Timer.h"

#include "ImageIO.h"
#include "ATIextensions.h"

#include "shader.h"

#define for if(0);else for

HAIL::Image2D *image;
HAIL::ImageFilter *filt1;
HAIL::ImageFilter *filt2;
HAIL::ImageFilter *filt3;

bool fft = false;
bool ifft = false;
bool apply1 = false;
bool apply2 = false;
bool apply3 = false;

char imageFile[256] = "../textures/mr_smiley.png";

Timer *time;
bool normalized = false;

bool displayStats = true;

char shaderFile[] = "nrm_display.txt";
GLuint nrmShader = 0;

Image screenShot;

//
//
////////////////////////////////////////////////////////////////////////////////
void keys( unsigned char c, int x, int y)
{
   static char buf[256];
   static int shot = 0;

   switch (c)
   {
      case '\033':
         delete image;
         HAIL::Shutdown();
         exit(0);
         break;
      case 'f':
         image->FFT();
         break;
      case 'i':
         image->IFFT();
         break;
      case '1':
         //gaussian blur
         image->ApplyFilter( filt1);
         break;
      case '2':
         //sharpen
         image->ApplyFilter( filt2);
         break;
      case '3':
         //Robert's edge detection
         image->ApplyFilter( filt3);
         break;
      case 'n':
         normalized = !normalized;
         break;
      case 's':
         imgGrabImage( screenShot, 0, 0, 512, 512, GL_RGBA, GL_UNSIGNED_BYTE);
         sprintf( buf, "shot%d.tga", shot++);
         imgSaveImage( screenShot, buf);
         break;
      case 'd':
         displayStats = ! displayStats;
         break;
   };
   glutPostRedisplay();
}

//
//
////////////////////////////////////////////////////////////////////////////////
void draw()
{
   static int which=0;
   glClear( GL_COLOR_BUFFER_BIT);
   

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0f, 1.0f, 0.0f, 1.0f);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glBindTexture( GL_TEXTURE_2D, image->GetTexHandle());
   glEnable(GL_TEXTURE_2D);

   if (normalized)
   {
      glEnable(GL_FRAGMENT_PROGRAM_ARB);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, nrmShader);
   }

   glColor3f( 1.0f, 1.0f, 1.0f);
   glBegin(GL_QUADS);
   glTexCoord2f( 0.0f, 0.0f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glTexCoord2f( 1.0f, 0.0f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glTexCoord2f( 1.0f, 1.0f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glTexCoord2f( 0.0f, 1.0f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable( GL_FRAGMENT_PROGRAM_ARB);

   if (displayStats)
   {
      glColor3f( 1.0f, 1.0f, 1.0f);
      glDisable(GL_TEXTURE_2D);
      time->UpdateTimer();
      char buffer[256];
      sprintf( buffer, "FPS: %3.2f", time->GetAvgHz());
      glColor3f( 1.0f, 0.0f, 0.0f);
      glRasterPos2f( 0.01f, 0.01f);
      for (char *c=buffer; *c != '\0'; c++)
      {
         glutBitmapCharacter( GLUT_BITMAP_TIMES_ROMAN_24, *c);
      }
   }

   glutSwapBuffers();
}

//
//
////////////////////////////////////////////////////////////////////////////////
void init()
{
   Image img;
   GLuint texID;
   float ref,x,y;

   img = imgCreateImage();
   if (imgLoadImage( img, imageFile))
   {
      printf( "Couldn't open image %s\n", imageFile);
   }
   glGenTextures( 1, &texID);
   glBindTexture( GL_TEXTURE_2D, texID);
   imgLoadImageAsTexture( img);
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE_EXT);
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE_EXT);
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
   image->LoadImageFromTexture( imgGetWidth(img), imgGetHeight(img), HAIL::RGBA, texID);


   //
   // Create a radius 3 gausian and load it as a separable filter
   float hor1[23];
   float vert1[23];
   float sigma = 0.0f;

   for (int ii=-11; ii<12; ii++)
   {
      hor1[ii+11] = expf(-(float)ii*(float)ii/(9.0f))/(2.0f*3.1415927f*9.0f);
      hor1[ii+11] = sqrtf(hor1[ii+11]);
      vert1[ii+11] = hor1[ii+11];
      sigma += hor1[ii+11];
   }

   filt1 = reinterpret_cast<HAIL::ImageFilter*>( new HAIL::Convolve2DSeperable( 23, 23, hor1, vert1));


   //
   //Create a simple sharpen filter [inseparable]
   float sharpen[] = {
      -1.0f, -1.0f, -1.0f, -1.0f, 9.0f, -1.0f, -1.0f, -1.0f, -1.0f
   };

   filt2 = reinterpret_cast<HAIL::ImageFilter*>( new HAIL::Convolve2D( 3, 3, sharpen));

   //
   //Create a Robert's edge detection filter
   float roberts[] = {
      0.0f, 0.0f, -1.0f, 0.0f, 1.0, 0.0f, 0.0f, 0.0f, 0.0f
   };

   filt3 = reinterpret_cast<HAIL::ImageFilter*>( new HAIL::Convolve2D( 3, 3, roberts));

   screenShot = imgCreateImage();
   
   nrmShader = loadFragmentProgram( shaderFile);
   
}

//
//
////////////////////////////////////////////////////////////////////////////////
void processCommandLine( int argc, char **argv)
{
   for (int ii=0; ii<argc; ii++)
   {
      if (!strcmp( argv[ii], "-t"))
      {
         strcpy( imageFile, argv[ii+1]);
         ii++;
      }
   }
}

//
//
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{
   glutInit(&argc, argv);
   glutInitWindowSize( 512, 512);
   processCommandLine( argc, argv);
   glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA);

   glutCreateWindow( "HW Image Processing");

   printf("HW image processing\n");
   printf("  [ESC] - quit program\n");
   printf("  'f'   - perform FFT (red channel only)\n");
   printf("  'i'   - perform IFFT (red/alpha to all channels)\n");
   printf("  '1'   - perform radius 3 gaussian blur\n");
   printf("  '2'   - perform a sharpening filter\n");
   printf("  '3'   - perform a Robert's edge filter\n\n");


   if (!HAIL::Init())
   {
      printf("Unable to allocate necessary resources\n");
      printf(" This program requires floating point textures and pbuffers\n");
   }

   image = new HAIL::Image2D;

   init();


   glutDisplayFunc(draw);
   glutIdleFunc(draw);
   glutKeyboardFunc(keys);

   time = new Timer();

   glutMainLoop();

   return 0;
}