//
// Image2D_FFT
//   Implementation for the FFT shaders and algorithms
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "HAIL.h"
#include "HAIL_private.h"

#include "Buffer.h"

#include "ATIExtensions.h"

using namespace HAIL;


#define for if(0);else for

//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::FFT()
{
   int pow2;


   BindBuffer( FLOAT_BUFFER);

   //really should recent the image to ensure it is a power of 2

   //
   //First, X dimension
   //

   //Configure transform
   glViewport( 0, 0, m_pow2Width, m_pow2Height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   //compute the power of 2 of the width
   pow2 = 0;
   while ( m_pow2Width > (1<<pow2) )
   {
      pow2++;
   }

   //perfrom scramble
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable( GL_TEXTURE_2D);
   glMatrixMode( GL_TEXTURE);
   glLoadIdentity();
   glMatrixMode( GL_MODELVIEW);

   //Setup the scamble texture
   // limit it to the correct mip level, as each level represents the transform
   // for a different power of 2
   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( SCRAMBLE_TEXTURE);
   glEnable( GL_TEXTURE_2D);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL_SGIS, 11-pow2);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL_SGIS, 11-pow2);

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hScramble);
   
   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   //Execute the scramble shader
   // texture coord 0 provides where in the whole image the pixel is
   // texture coord 1 provides where in the scramble texture to get the offset value 
   glBegin( GL_QUADS);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
      glVertex3f( 0.0f, 0.0f, -0.5f);

      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
      glVertex3f( 1.0f, 0.0f, -0.5f);

      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
      glVertex3f( 1.0f, 1.0f, -0.5f);

      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
      glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
   

   //perform butterflies   

   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( static_cast<UtilityTexture>(FFT_1024_TEXTURE + (10-pow2)*2));
   glEnable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hButterfly);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   //Perform log2(width) passes of butterflies
   for (int ii=0; ii<pow2; ii++)
   {
      //perfrom butterfly by applying the shader to a single quad
      // the texture coords on unit one change to select the correct
      // offsets and factor
      glBegin( GL_QUADS);

         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
         glVertex3f( 0.0f, 0.0f, -0.5f);
      
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
         glVertex3f( 1.0f, 0.0f, -0.5f);
      
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
         glVertex3f( 1.0f, 1.0f, -0.5f);
      
         glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
         glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
         glVertex3f( 0.0f, 1.0f, -0.5f);
      
      glEnd();

      //copy out the image to prepare for the next pass
      glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
   }

   glDisable(GL_FRAGMENT_PROGRAM_ARB);   
  
   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glActiveTextureARB( GL_TEXTURE0_ARB);

   //
   //Next, Y dimension
   //

   //compute the power of two
   pow2 = 0;
   while ( m_pow2Height > (1<<pow2) )
   {
      pow2++;
   }

   //scramble
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( SCRAMBLE_TEXTURE);
   glEnable( GL_TEXTURE_2D);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL_SGIS, 11-pow2);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL_SGIS, 11-pow2);
   


   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vScramble);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);

   //perform butterflies
   

   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( static_cast<UtilityTexture>(FFT_1024_TEXTURE + (10-pow2)*2));
   
   glEnable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vButterfly);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   for (int ii=0; ii<pow2; ii++)
   {
      
      glBegin( GL_QUADS);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 1.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 1.0f, -0.5f);
      
      glEnd();

      glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      
   }

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);

   glDisable(GL_FRAGMENT_PROGRAM_ARB);


   UnbindBuffer(FLOAT_BUFFER);
}

//
//
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::IFFT()
{
   int pow2;

   
   

   BindBuffer( FLOAT_BUFFER);

   //Configure transform
   glViewport( 0, 0, m_pow2Width, m_pow2Height);
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   

   //
   //First, X dimension
   //

   pow2 = 0;
   while ( m_pow2Width > (1<<pow2) )
   {
      pow2++;
   }

   //perfrom scramble
   glActiveTextureARB( GL_TEXTURE0_ARB);
   glBindTexture( GL_TEXTURE_2D, m_fTexHandle);
   glEnable( GL_TEXTURE_2D);
   glMatrixMode( GL_TEXTURE);
   glLoadIdentity();
   glMatrixMode( GL_MODELVIEW);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( SCRAMBLE_TEXTURE);
   glEnable( GL_TEXTURE_2D);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL_SGIS, 11-pow2);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL_SGIS, 11-pow2);
   

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hScramble);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
   

   //perform butterflies
   

   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( static_cast<UtilityTexture>(IFFT_1024_TEXTURE + (10-pow2)*2));
   glEnable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   
   

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hButterfly);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);


   for (int ii=0; ii<pow2; ii++)
   {
      
      glBegin( GL_QUADS);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 1.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 1.0f, -0.5f);
      
      glEnd();

      glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      
   }

   glDisable(GL_FRAGMENT_PROGRAM_ARB);   
  
   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glActiveTextureARB( GL_TEXTURE0_ARB);

   //
   //Next, Y dimension
   //

   pow2 = 0;
   while ( m_pow2Height > (1<<pow2) )
   {
      pow2++;
   }

   //scramble
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( SCRAMBLE_TEXTURE);
   glEnable( GL_TEXTURE_2D);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL_SGIS, 11-pow2);
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL_SGIS, 11-pow2);

   

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vScramble);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, 0.5f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, 0.5f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);

   //perform butterflies
   

   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glActiveTextureARB( GL_TEXTURE1_ARB);
   BindUtilityTexture( static_cast<UtilityTexture>(IFFT_1024_TEXTURE + (10-pow2)*2));
   glEnable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);
   


   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vButterfly);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

 
   for (int ii=0; ii<pow2; ii++)
   {
      
      glBegin( GL_QUADS);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 0.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 0.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 1.0f, 1.0f, -0.5f);
      glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
      glMultiTexCoord2fARB( GL_TEXTURE1_ARB, 1.0f, ii/16.0f + 1.0f/32.0f);
      glVertex3f( 0.0f, 1.0f, -0.5f);
      
      glEnd();

      glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);
      
   }

   glActiveTextureARB( GL_TEXTURE1_ARB);
   glBindTexture( GL_TEXTURE_2D, 0);
   glDisable( GL_TEXTURE_2D);

   glActiveTextureARB( GL_TEXTURE0_ARB);

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

  
   

   //remap
   glMatrixMode( GL_PROJECTION);
   glLoadIdentity();
   gluOrtho2D( 0.0, 1.0, 0.0, 1.0);
   glMatrixMode( GL_MODELVIEW);
   glLoadIdentity();

   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::rescale);
   float temp = 1.0f/(m_pow2Width*m_pow2Height);

   glProgramLocalParameter4fARB( GL_FRAGMENT_PROGRAM_ARB, 0, temp, temp, temp, temp);

   glEnable(GL_FRAGMENT_PROGRAM_ARB);

   glBegin( GL_QUADS);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 0.0f);
   glVertex3f( 0.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 0.0f);
   glVertex3f( 1.0f, 0.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 1.0f, 1.0f);
   glVertex3f( 1.0f, 1.0f, -0.5f);
   glMultiTexCoord2fARB( GL_TEXTURE0_ARB, 0.0f, 1.0f);
   glVertex3f( 0.0f, 1.0f, -0.5f);
   glEnd();

   glDisable(GL_FRAGMENT_PROGRAM_ARB);

   glCopyTexSubImage2D( GL_TEXTURE_2D, 0, 0, 0, 0, 0, m_pow2Width, m_pow2Height);

   UnbindBuffer(FLOAT_BUFFER);
}

//
//
//
//////////////////////////////////////////////////////////////////////////////////////////
unsigned int Image2D::hScramble = 0;
unsigned int Image2D::vScramble = 0;
unsigned int Image2D::hButterfly = 0;
unsigned int Image2D::vButterfly = 0;
unsigned int Image2D::rescale = 0;

//
//InitFFTShader
//
//  Used to precompile the shaders used for FFT/IFFT.
//////////////////////////////////////////////////////////////////////////////////////////
void Image2D::InitFFTShaders()
{
   

   if (Image2D::hScramble)
   {
      return;
   }

   BindBuffer(FLOAT_BUFFER);

   {
      strstream shader;
      //
      //create horizontal scramble shader
      //
      CreateShaderPreamble( shader, 2, 1);

      shader << "TEX r0, t1, texture[1], 2D;\n";
      shader << "MOV r0.y, t0;\n";
      shader << "TEX r0, r0, texture[0], 2D;\n";
      shader << "MOV res, r0;\n";
      shader << "END\n\n" << '\0';


      glGenProgramsARB( 1, &Image2D::hScramble);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hScramble);
      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());

      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      }
   }

   
   {
      strstream shader;
      //
      //create the horizontal butterfly shader
      //
      CreateShaderPreamble( shader, 2, 4);

      //fetch the offsets and factor
      shader << "TEX r0, t1, texture[1], 2D;\n";
   
      //configure the texture coords, selecting the correct x offsets
      // tha ABS instructions are used becuase the sign encodes extra info
      shader << "MOV r1.y, t0.y;\n";                
      shader << "ABS r1.x, r0.x;\n";
      shader << "MOV r2.y, t0.y;\n";
      shader << "ABS r2.x, r0.y;\n";

      //obtain f[i] and f[j]
      shader << "TEX r1, r1, texture[0], 2D;\n";
      shader << "TEX r2, r2, texture[0], 2D;\n";

      // f[j]*W, complex multiply
      shader << "MUL r3.rgb, r0.a, r1.a;\n";
      shader << "MAD r3.rgb, r0.b, r1.r, -r3.r;\n";
      shader << "MUL r3.a, r0.b, r1.a;\n";
      shader << "MAD r3.a, r0.a, r1.r, r3.a;\n";

      //determine if this is i or j
      shader << "PARAM c0 = 1.0;\n";
      shader << "CMP r0.x, r0.x, -c0.x, c0.x;\n";

      //add it to complete the butterfly
      // f[i] = f[i] + f[j]*w
      // f[j] = f[i] - f[j]*w
      shader << "MAD res, r0.x, r3, r2;\n";

      shader << "END\n\n" << '\0';

      glGenProgramsARB( 1, &Image2D::hButterfly);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::hButterfly);
      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());

      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      }
   }

   {
      strstream shader;

      //
      //create scramble shader for y dimension
      // shader is very simple, only using the scramble texture to twiddle
      // the y component
      //
      CreateShaderPreamble( shader, 2, 1);
      shader << "TEX r0, t1, texture[1], 2D;\n";
      shader << "MOV r0.x, t0;\n";
      shader << "TEX r0, r0, texture[0], 2D;\n";
      shader << "MOV res, r0;\n";
      shader << "END\n\n" << '\0';

      glGenProgramsARB( 1, &Image2D::vScramble);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vScramble);
      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());

      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      }
   }

   {
      strstream shader;

      //
      //create vertical butterfly shader
      //
      CreateShaderPreamble( shader, 2, 4);

      //declare params

      //fetch the offsets and factor
      shader << "TEX r0, t1, texture[1], 2D;\n";
   
      //configure the texture coords
      shader << "MOV r1.x, t0.x;\n";                
      shader << "ABS r1.y, r0.x;\n";
      shader << "MOV r2.x, t0.x;\n";
      shader << "ABS r2.y, r0.y;\n";

      //obtain f[i] and f[j]
      shader << "TEX r1, r1, texture[0], 2D;\n";
      shader << "TEX r2, r2, texture[0], 2D;\n";

      // f[j]*W
      shader << "MUL r3.rgb, r0.a, r1.a;\n";
      shader << "MAD r3.rgb, r0.b, r1.r, -r3.r;\n";
      shader << "MUL r3.a, r0.b, r1.a;\n";
      shader << "MAD r3.a, r0.a, r1.r, r3.a;\n";

      //determine if this is i or j
      shader << "PARAM c0 = 1.0;\n";
      shader << "CMP r0.x, r0.x, -c0.x, c0.x;\n";

      //add it to complete the butterfly
      shader << "MAD res, r0.x, r3, r2;\n";

      shader << "END\n\n" << '\0';

      glGenProgramsARB( 1, &Image2D::vButterfly);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::vButterfly);
      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());

      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      }
   }

   {
      //
      // create rescale shader used for IFFT
      //
      strstream shader;

      CreateShaderPreamble( shader, 1, 1);
      shader << "TEX r0, t0, texture[0], 2D;\n";
      shader << "PARAM c0 = program.local[0];\n";
  
      shader << "MUL res, r0.r, c0;\n";
      shader << "END\n\n" << '\0';


      glGenProgramsARB( 1, &Image2D::rescale);
      glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::rescale);
      glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(shader.str()), shader.str());

      if ( glGetError())
      {
         const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
         int pos;
         glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      }
   }

   UnbindBuffer(FLOAT_BUFFER);
}