//
// Image2D_shaders.cpp
//   Implementation of utility shaders for Image2D class
//
//
//
//
//
//
////////////////////////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include <stdio.h>
#include <string.h>

#include "HAIL.h"
#include "HAIL_private.h"

#include "Buffer.h"

#include "ATIExtensions.h"

using namespace HAIL;


#define for if(0);else for

//
//
//
////////////////////////////////////////////////////////////////////////////////
unsigned int Image2D::nearest = 0;
unsigned int Image2D::linear = 0;
unsigned int Image2D::cubic = 0;
unsigned int Image2D::add = 0;
unsigned int Image2D::subtract = 0;
unsigned int Image2D::modulate = 0;
unsigned int Image2D::replace = 0;

//
//
//
////////////////////////////////////////////////////////////////////////////////
void Image2D::InitInterpolationShaders()
{

   if (Image2D::nearest)
   {
      return;
   }

   BindBuffer(FLOAT_BUFFER);
   strstream nearest;
   strstream linear;
   strstream cubic;

   //Nearest Interpolation 
   CreateShaderPreamble( nearest, 3, 7);
   //filter coefficients
   nearest << "TEX r0, t2, texture[0], 2D;\n";
   nearest << "MOV res, r0;\n";
   nearest << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::nearest);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::nearest);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(nearest.str()), nearest.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf( "## Error in nearest sample shader ##\n");
      printf( "  %s\n\n", dbg);
   }

   CreateShaderPreamble( linear, 3, 7);
   //filter coefficients
   linear << "TEX r0, t0, texture[1], 2D;\n";
   linear << "TEX r1, t1, texture[1], 2D;\n";
   //create tex coord offsets
   linear << "PARAM c0 = program.local[0];\n";
   linear << "PARAM c1 = { -0.5, 0.5, -0.5};\n";
   linear << "MUL r6.ra, r0, r1;\n";
   linear << "MUL r6.g, r0.r, r1.a;\n";
   linear << "MUL r6.b, r0.a, r1.r;\n";
   linear << "MAD r2.xy, c0, c1.g, t2;\n";
   linear << "MAD r3.xy, c0, c1.gbra, t2;\n";
   linear << "MAD r4.xy, c0, c1.rgba, t2;\n";
   linear << "MAD r5.xy, c0, c1.r, t2;\n";
   linear << "TEX r2, r2, texture[0], 2D;\n";
   linear << "TEX r3, r3, texture[0], 2D;\n";
   linear << "TEX r4, r4, texture[0], 2D;\n";
   linear << "TEX r5, r5, texture[0], 2D;\n";
   linear << "MUL r2, r2, r6.r;\n";
   linear << "MAD r2, r4, r6.g, r2;\n";
   linear << "MAD r2, r3, r6.b, r2;\n";
   linear << "MAD res, r5, r6.a, r2;\n";
   linear << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::linear);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::linear);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(linear.str()), linear.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf( "## Error in linear sample shader ##\n");
      printf( "  %s\n\n", dbg);
   }

   CreateShaderPreamble( cubic, 3, 12);
   //filter coefficients
   cubic << "TEX r0, t0, texture[1], 2D;\n";
   cubic << "TEX r1, t1, texture[1], 2D;\n";
   //create tex coord offsets
   cubic << "PARAM c0 = program.local[0];\n";
   cubic << "PARAM c1 = { -1.5, -0.5, 0.5, 1.5};\n";

   //cubic << "ADD r2.x, t2, c0.a;\n";
   //cubic << "ADD r2.y, t2, c1.r;\n";
   /* cubic << "MAD r2.xy, c0, c1.arrb, t2;\n"; */
   cubic << "MAD r2.x, c0, c1.a, t2;\n";
   cubic << "MAD r2.y, c0, c1.r, t2;\n";
   //cubic << "ADD r3.x, t2, c0.b;\n";
   //cubic << "ADD r3.y, t2, c1.r;\n";
   cubic << "MAD r3.xy, c0, c1.brga, t2;\n";
   //cubic << "ADD r4.x, t2, c0.g;\n";
   //cubic << "ADD r4.y, t2, c1.r;\n";
   /* cubic << "MAD r4.xy, c0, c1.grra, t2;\n"; */
   cubic << "MAD r4.x, c0, c1.g, t2;\n";
   cubic << "MAD r4.y, c0, c1.r, t2;\n";
   //cubic << "ADD r5.x, t2, c0.r;\n";
   //cubic << "ADD r5.y, t2, c1.r;\n";
   cubic << "MAD r5.xy, c0, c1.rrrr, t2;\n";
   //cubic << "ADD r6.x, t2, c0.a;\n";
   //cubic << "ADD r6.y, t2, c1.g;\n";
   /*  cubic << "MAD r6.xy, c0, c1.aggr, t2;\n"; */
   cubic << "MAD r6.x, c0, c1.a, t2;\n";
   cubic << "MAD r6.y, c0, c1.g, t2;\n";
   //cubic << "ADD r7.x, t2, c0.b;\n";
   //cubic << "ADD r7.y, t2, c1.g;\n";
   cubic << "MAD r7.xy, c0, c1.bgra, t2;\n";
   //cubic << "ADD r8.x, t2, c0.g;\n";
   //cubic << "ADD r8.y, t2, c1.g;\n";
   cubic << "MAD r8.xy, c0, c1.gggg, t2;\n";
   //cubic << "ADD r9.x, t2, c0.r;\n";
   //cubic << "ADD r9.y, t2, c1.g;\n";
   cubic << "MAD r9.xy, c0, c1.rgba, t2;\n";
   
   cubic << "TEX r2, r2, texture[0], 2D;\n";
   cubic << "TEX r3, r3, texture[0], 2D;\n";
   cubic << "TEX r4, r4, texture[0], 2D;\n";
   cubic << "TEX r5, r5, texture[0], 2D;\n";
   cubic << "TEX r6, r6, texture[0], 2D;\n";
   cubic << "TEX r7, r7, texture[0], 2D;\n";
   cubic << "TEX r8, r8, texture[0], 2D;\n";
   cubic << "TEX r9, r9, texture[0], 2D;\n";

   //accumulate the samples
   cubic << "MUL r10, r0, r1.a;\n";
   cubic << "MUL r11, r2, r10.r;\n";
   cubic << "MAD r11, r3, r10.g, r11;\n";
   cubic << "MAD r11, r4, r10.b, r11;\n";
   cubic << "MAD r11, r5, r10.a, r11;\n";
   cubic << "MUL r10, r0, r1.b;\n";
   cubic << "MAD r11, r6, r10.r, r11;\n";
   cubic << "MAD r11, r7, r10.g, r11;\n";
   cubic << "MAD r11, r8, r10.b, r11;\n";
   cubic << "MAD r11, r9, r10.a, r11;\n";

   //cubic << "ADD r2.x, t2, c0.a;\n";
   //cubic << "ADD r2.y, t2, c1.b;\n";
   cubic << "MAD r2.xy, c0, c1.abgr, t2;\n";
   //cubic << "ADD r3.x, t2, c0.b;\n";
   //cubic << "ADD r3.y, t2, c1.b;\n";
   cubic << "MAD r3.xy, c0, c1.bbbb, t2;\n";
   //cubic << "ADD r4.x, t2, c0.g;\n";
   //cubic << "ADD r4.y, t2, c1.b;\n";
   cubic << "MAD r4.xy, c0, c1.gbra, t2;\n";
   //cubic << "ADD r5.x, t2, c0.r;\n";
   //cubic << "ADD r5.y, t2, c1.b;\n";
   /* cubic << "MAD r5.xy, c0, c1.rbba, t2;\n"; */
   cubic << "MAD r5.x, c0, c1.r, t2;\n";
   cubic << "MAD r5.y, c0, c1.b, t2;\n";
   //cubic << "ADD r6.x, t2, c0.a;\n";
   //cubic << "ADD r6.y, t2, c1.a;\n";
   cubic << "MAD r6.xy, c0, c1.aaaa, t2;\n";
   //cubic << "ADD r7.x, t2, c0.b;\n";
   //cubic << "ADD r7.y, t2, c1.a;\n";
   /* cubic << "MAD r7.xy, c0, c1.baar, t2;\n"; */
   cubic << "MAD r7.x, c0, c1.b, t2;\n";
   cubic << "MAD r7.y, c0, c1.a, t2;\n";
   //cubic << "ADD r8.x, t2, c0.g;\n";
   //cubic << "ADD r8.y, t2, c1.a;\n";
   /* cubic << "MAD r8.xy, c0, c1.gaar, t2;\n"; */
   cubic << "MAD r8.x, c0, c1.g, t2;\n";
   cubic << "MAD r8.y, c0, c1.a, t2;\n";
   //cubic << "ADD r9.x, t2, c0.r;\n";
   //cubic << "ADD r9.y, t2, c1.a;\n";
   /* cubic << "MAD r9.xy, c0, c1.raab, t2;\n"; */
   cubic << "MAD r9.x, c0, c1.r, t2;\n";
   cubic << "MAD r9.y, c0, c1.a, t2;\n";

   cubic << "TEX r2, r2, texture[0], 2D;\n";
   cubic << "TEX r3, r3, texture[0], 2D;\n";
   cubic << "TEX r4, r4, texture[0], 2D;\n";
   cubic << "TEX r5, r5, texture[0], 2D;\n";
   cubic << "TEX r6, r6, texture[0], 2D;\n";
   cubic << "TEX r7, r7, texture[0], 2D;\n";
   cubic << "TEX r8, r8, texture[0], 2D;\n";
   cubic << "TEX r9, r9, texture[0], 2D;\n";

   cubic << "MUL r10, r0, r1.g;\n";
   cubic << "MAD r11, r2, r10.r, r11;\n";
   cubic << "MAD r11, r3, r10.g, r11;\n";
   cubic << "MAD r11, r4, r10.b, r11;\n";
   cubic << "MAD r11, r5, r10.a, r11;\n";
   cubic << "MUL r10, r0, r1.r;\n";
   cubic << "MAD r11, r6, r10.r, r11;\n";
   cubic << "MAD r11, r7, r10.g, r11;\n";
   cubic << "MAD r11, r8, r10.b, r11;\n";
   cubic << "MAD res, r9, r10.a, r11;\n";
   cubic << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::cubic);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::cubic);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(cubic.str()), cubic.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf( "## Error in cubic sample shader ##\n");
      printf( "  %s\n\n", dbg);
   }

   UnbindBuffer(FLOAT_BUFFER);
}

//
//
//
////////////////////////////////////////////////////////////////////////////////
void Image2D::InitApplyShaders()
{
   if ( Image2D::add)
      return;

   BindBuffer(FLOAT_BUFFER);

   strstream add;
   strstream subtract;
   strstream modulate;
   strstream replace;

   CreateShaderPreamble( add, 2, 2);
   add << "TEX r0, t0, texture[0], 2D;\n";
   add << "TEX r1, t1, texture[1], 2D;\n";
  
   add << "ADD res, r0, r1;\n\n";
         
   add << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::add);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::add);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(add.str()), add.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf("# Error failed to load add shader #\n");
      printf(" -%s\n", dbg);
   }

   CreateShaderPreamble( modulate, 2, 2);
   modulate << "TEX r0, t0, texture[0], 2D;\n";
   modulate << "TEX r1, t1, texture[1], 2D;\n";
   
   modulate << "MUL res, r0, r1;\n\n";
         
   modulate << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::modulate);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::modulate);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(modulate.str()), modulate.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf("# Error failed to load modulate shader #\n");
      printf(" -%s\n", dbg);
   }

   CreateShaderPreamble( subtract, 2, 2);
   subtract << "TEX r0, t0, texture[0], 2D;\n";
   subtract << "TEX r1, t1, texture[1], 2D;\n";

   subtract << "ADD res, r0, -r1;\n\n";
         
   subtract << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::subtract);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::subtract);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(subtract.str()), subtract.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf("# Error failed to load subtract shader #\n");
      printf(" -%s\n", dbg);
   }

   CreateShaderPreamble( replace, 2, 2);
   replace << "TEX r0, t0, texture[0], 2D;\n";
   replace << "TEX r1, t1, texture[1], 2D;\n";
   
   replace << "MOV res, r1;\n\n";

   replace << "END\n\n" << '\0';

   glGenProgramsARB( 1, &Image2D::replace);
   glBindProgramARB( GL_FRAGMENT_PROGRAM_ARB, Image2D::replace);
   glProgramStringARB( GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(replace.str()), replace.str());

   if ( glGetError())
   {
      const GLubyte *dbg = glGetString(GL_PROGRAM_ERROR_STRING_ARB);
      int pos;
      glGetIntegerv( GL_PROGRAM_ERROR_POSITION_ARB, &pos);
      printf("# Error failed to load replace shader #\n");
      printf(" -%s\n", dbg);
   }

   UnbindBuffer(FLOAT_BUFFER);
}