#include "Model.h"
#include "../include/3ds.h"

Model::Model(){
}

Model::~Model(){
	for (unsigned int i = 0; i < meshes.getSize(); i++){
		if (meshes[i].vertices != NULL) delete meshes[i].vertices;
		//if (vaoBuffer != 0) glFreeObjectBufferATI(vaoBuffer);
	}
}


bool Model::loadFile(char *fileName, const int x, const int y, const int z, const float scale){
	char *ext = strrchr(fileName, '.');
	if (ext != NULL && strcmp(ext + 1, "3ds") == 0){
		data3ds_t model;
		int i,j,k;

		Read3dsFile(fileName, &model);
		Calculate3dsNormals(&model);
		//Calculate3dsTangentSpace(&model);

		for (i = 0; i < model.meshCount; i++){
			Mesh mesh;

			mesh.nVertices = model.meshes[i].triCount * 3;
			mesh.vertices = new ModelVertex[mesh.nVertices];
			ModelVertex *dest = mesh.vertices;
			for (j = 0; j < model.meshes[i].triCount; j++){
				for (k = 0; k < 3; k++){
					long vert = model.meshes[i].tris[j][k];
					
					dest->vertex.x = model.meshes[i].verts[vert][x] * scale;
					dest->vertex.y = model.meshes[i].verts[vert][y] * scale;
					dest->vertex.z = model.meshes[i].verts[vert][z] * scale;
					
					dest->normal.x = model.meshes[i].norms[3 * j + k][x];
					dest->normal.y = model.meshes[i].norms[3 * j + k][y];
					dest->normal.z = model.meshes[i].norms[3 * j + k][z];
					dest++;
				}
			}

			// Assuming just one group per mesh
			memcpy(mesh.diffuse, model.materials[model.meshes[i].groups[0].mat].diffuse, 4 * sizeof(float));

			meshes.add(mesh);
		}

/*		for (i = 0; i < model.materialCount; i++){
			printf("%s\n", model.materials[i].texture);
			printf("%.4f %.4f %.4f\n", model.materials[i].diffuse[0], model.materials[i].diffuse[1], model.materials[i].diffuse[2]);
			printf("%.4f %.4f %.4f\n", model.materials[i].ambient[0], model.materials[i].ambient[1], model.materials[i].ambient[2]);
			printf("%.4f %.4f %.4f\n", model.materials[i].specular[0], model.materials[i].specular[1], model.materials[i].specular[2]);

			printf("\n");
		}
*/
      min[0] = model.min[0];
      min[1] = model.min[1];
      min[2] = model.min[2];

      max[0] = model.max[0];
      max[1] = model.max[1];
      max[2] = model.max[2];

		Free3dsData(&model);
	} else {

		return false;
	}

	/*if (GL_ATI_vertex_array_object){
		vaoBuffer = glNewObjectBufferATI(nVertices * sizeof(ModelVertex), vertices, GL_STATIC_ATI);
	}*/

	return true;
}

void Model::draw(){
	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	for (unsigned int i = 0; i < meshes.getSize(); i++){
		/*if (GL_ATI_vertex_array_object){
			glArrayObjectATI(GL_VERTEX_ARRAY, 3, GL_FLOAT, sizeof(ModelVertex), vaoBuffer, 0);
			glArrayObjectATI(GL_NORMAL_ARRAY, 3, GL_FLOAT, sizeof(ModelVertex), vaoBuffer, 3 * sizeof(float));
		} else {*/
			glVertexPointer(3, GL_FLOAT, sizeof(ModelVertex), meshes[i].vertices);
			glNormalPointer(GL_FLOAT, sizeof(ModelVertex), ((char *) meshes[i].vertices) + 3 * sizeof(float));
		//}

		glColor4fv(meshes[i].diffuse);

		glDrawArrays(GL_TRIANGLES, 0, meshes[i].nVertices);

	}

	glDisableClientState(GL_NORMAL_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
}

void Model::drawDepth(){
	glEnableClientState(GL_VERTEX_ARRAY);

	for (unsigned int i = 0; i < meshes.getSize(); i++){
		glVertexPointer(3, GL_FLOAT, sizeof(ModelVertex), meshes[i].vertices);

		glDrawArrays(GL_TRIANGLES, 0, meshes[i].nVertices);
	}

	glDisableClientState(GL_VERTEX_ARRAY);
}

float Model::diameter() {
   float val[3];

   val[0] = max[0] - min[0];
   val[1] = max[1] - min[1];
   val[2] = max[2] - min[2];

   return sqrtf( val[0]*val[0] + val[1]*val[1] + val[2]*val[2]);
}
