/*
	Emil Persson
	epersson@ati.com
	ATI Technologies Inc.

	A set of utility functions for common stuff
*/

#include <GL/glut.h>


// Using display lists to keep overhead down.
// Without display lists the overhead of the glutBitmapCharacter()
// is high enough to clearly reduce the framerate
static GLuint charDLs[127];

void drawText(char *string, const float x, const float y){
	if (charDLs[0] == 0){
		for (unsigned int i = 0; i < 127; i++){
			charDLs[i] = glGenLists(1);
			glNewList(charDLs[i], GL_COMPILE);
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, i + 1);
			glEndList();
		}
	}

	glRasterPos2f(x, y);
	
	while (*string){
		if (*string > 0) glCallList(charDLs[(*string) - 1]);
		string++;
	}
}

void drawTextureText(char *string, float x, float y, const float w, const float h, const int nX, const int nY){
    float tx,ty;
	float sx = x;

	glBegin(GL_QUADS);
    while (*string){
        if (*string == '\n'){
            y += h;
            x = sx;
        } else {
            tx = float(((unsigned char) *string) % nX) / nX;
            ty = float(((unsigned char) *string) / nX) / nY;

            //tx = float(((unsigned char) *string) & 15) / 16;
            //ty = float(((unsigned char) *string) >> 4) / 16;

			glTexCoord2f(tx, ty);
			glVertex2f(x, y);
			
			glTexCoord2f(tx + 1.0f / nX, ty);
			glVertex2f(x + w, y);
			
			glTexCoord2f(tx + 1.0f / nX, ty + 1.0f / nY);
			glVertex2f(x + w, y + h);
			
			glTexCoord2f(tx, ty + 1.0f / nY);
			glVertex2f(x, y + h);

            x += w;
        }
        string++;
    }
    glEnd();
}
