/*
	Emil Persson
	epersson@ati.com
	ATI Technologies Inc.

	The Vector class implements the most common vector operations and
	overloads the standard mathematical operators so vectors can be
	used the same way as you would on paper.
*/

#include "Vector.h"
#include <math.h>

void Vector::operator += (const Vector &v){
	x += v.x;
	y += v.y;
	z += v.z;
}

void Vector::operator -= (const Vector &v){
	x -= v.x;
	y -= v.y;
	z -= v.z;
}

Vector operator + (const Vector &u, const Vector &v){
	return Vector(u.x + v.x,u.y + v.y,u.z + v.z);
}

Vector operator - (const Vector &u, const Vector &v){
	return Vector(u.x - v.x,u.y - v.y,u.z - v.z);
}

Vector operator - (const Vector &v){
	return Vector(-v.x,-v.y,-v.z);
}

 // This is a dot product and not a per-component multiplication
float operator * (const Vector &u, const Vector &v){
	return u.x * v.x + u.y * v.y + u.z * v.z;
}

Vector operator * (const float scalar, const Vector &v){
	return Vector(scalar * v.x,scalar * v.y,scalar * v.z);
}

Vector operator * (const Vector &v, const float scalar){
	return Vector(scalar * v.x,scalar * v.y,scalar * v.z);
}


Vector operator / (const Vector &v, const float dividend){
	return Vector(v.x / dividend,v.y / dividend,v.z / dividend);
}

Vector cross(const Vector &u, const Vector &v){
	return Vector(u.y * v.z - v.y * u.z, u.z * v.x - u.x * v.z, u.x * v.y - u.y * v.x);
}

float length(const Vector &v){
	return sqrtf(v.x * v.x + v.y * v.y + v.z * v.z);
}

float lengthSqr(const Vector &v){
	return v.x * v.x + v.y * v.y + v.z * v.z;
}

Vector normalize(const Vector &v){
	float invLen = 1.0f / sqrtf(v.x * v.x + v.y * v.y + v.z * v.z);
	return v * invLen;
}
