/*
	Emil Persson
	epersson@ati.com
	ATI Technologies Inc.

	The Vector class implements the most common vector operations and
	overloads the standard mathematical operators so vectors can be
	used the same way as you would on paper.
*/

#ifndef _VECTOR_H_
#define _VECTOR_H_

struct Vector {
	float x,y,z;
	Vector(){}
	Vector(const float v[3]){
		x = v[0];
		y = v[1];
		z = v[2];
	}
	Vector(const float ix, const float iy, const float iz){
		x = ix;
		y = iy;
		z = iz;
	}
	operator const float *() const {
		return &x;
	}

	void operator += (const Vector &v);
	void operator -= (const Vector &v);
};

Vector operator + (const Vector &u, const Vector &v);
Vector operator - (const Vector &u, const Vector &v);
Vector operator - (const Vector &v);
 // This is a dot product and not a per-component multiplication
float  operator * (const Vector &u, const Vector &v);
Vector operator * (const float scalar, const Vector &v);
Vector operator * (const Vector &v, const float scalar);
Vector operator / (const Vector &v, const float dividend);

Vector cross(const Vector &u, const Vector &v);
float length(const Vector &v);
float lengthSqr(const Vector &v);
Vector normalize(const Vector &v);

#endif
